<?php

namespace Leenooks\SQRL;

use Illuminate\Support\Arr;
use SimpleSoftwareIO\QrCode\Generator as QRCode;

use Leenooks\SQRL\SQRL\Nonce;
use Leenooks\SQRL\SQRL\Pubkey;

class SQRL
{
	//TIF Codes
	protected static $tif_codes = [
		'ID_MATCH'				=> 0x01,
		'PREV_ID_MATCH'			=> 0x02,
		'IP_MATCH'				=> 0x04,
		'SQRL_DISABLED'			=> 0x08,
		'UNSUPPORTED_FUNCTION'	=> 0x10,
		'TRANSIENT_ERROR'		=> 0x20,
		'COMMAND_FAILED'		=> 0x40,
		'CLIENT_FAILURE'		=> 0x80,
		'BAD_ID'				=> 0x100,
		'ID_SUPERSEDED'			=> 0x200,
	];

	/**
	 * Base 64 Decode a URL encoded string
	 *
	 * @param string $url
	 * @return string
	 */
	public static function base64_decode_url(string $url): string
	{
		return base64_decode(str_replace(['-','_'],['+','/'],$url));
	}

	/**
	 * Base 64 Encode a URL
	 *
	 * @param string $url
	 * @return string|string[]
	 */
	public static function base64_encode_url(string $url): string
	{
		return str_replace(['+','/','='],['-','_',''],base64_encode($url));
	}

	/**
	 * Return a nonce used for authentication
	 *
	 * @return array
	 */
	public static function authNonce(int $size=100): array
	{
		$url = config('sqrl.url');

		$o = Nonce::createAuth('',$url);
		$o->url = self::base64_encode_url(sprintf('%s?nut=%s',$url,$o->nonce));
		$o->save();

		$route = sprintf('%s?nut=%s',config('sqrl.api_route'),$o->nonce);
		$sqrl_url = sprintf('sqrl://%s',config('sqrl.domain').$route);

		return [
			'nonce'=>$o->nonce,
			'check_state_on'=>$route,
			'sqrl_url'=>$sqrl_url,
			'sqrl_url_encoded'=>self::base64_encode_url(sprintf('%s&can=%s',$sqrl_url,$o->can)),
			'qrcode'=>SQRL::qrCode($sqrl_url,$size),
		];
	}

	protected static function clientDecode(string $clientInput): array
	{
		$inputAsArray = explode("\n",self::base64_decode_url($clientInput));
		$result = collect();

		foreach (array_filter($inputAsArray) as $individualInputs) {
			if (strpos($individualInputs,'=') === FALSE) {
				continue;
			}

			list($key,$val) = explode("=", $individualInputs);
			$val = trim($val);		// Trim \r

			switch ($key) {
				case 'cmd':
				case 'btn':
				case 'ver':
					$result->put($key,$val);
					break;

				case 'idk':
				case 'ins':
				case 'pidk':
				case 'suk':
				case 'vuk':
					$result->put($key,self::base64_decode_url($val));
					break;

				case 'opt':
					$result->put($key,explode('~',$val));
					break;
			}

		}

		return $result->toArray();
	}

	public static function decodeData(array $data_encoded): array
	{
		$result = collect();

		if (($x=Arr::get($data_encoded,'client'))) {
			$result->put('client',self::clientDecode($x));
		}

		if (($x=Arr::get($data_encoded,'server'))) {
			$result->put('server',self::serverDecode($x));
		}

		foreach (['ids','urs','pids'] as $key) {
			if (($x=Arr::get($data_encoded,$key))) {
				$result->put($key,self::base64_decode_url($x));
			}
		}

		return $result->toArray();
	}

	protected static function serverDecode(string $serverData)
	{
		$decoded = self::base64_decode_url($serverData);

		if ((substr($decoded,0,7) === 'sqrl://') || (substr($decoded,0,6) === 'qrl://')) {
			return $decoded;
		}

		$parsedResult = collect();
		$serverValues = explode("\r\n",$decoded);

		foreach ($serverValues as $value) {
			$splitStop = strpos($value,'=');

			$key = substr($value,0,$splitStop);
			$val = substr($value,$splitStop+1);

			$parsedResult->put($key,$val);
		}

		return $parsedResult->toArray();
	}

	/**
	 * Return a QRCode image
	 *
	 * @param string $nonce
	 * @param int $size
	 * @return string
	 */
	protected static function qrCode(string $nonce,int $size=100): string
	{
		return (new QRCode)->size($size)->generate($nonce);
	}

	/**
	 * Get the TIF Code
	 *
	 * @param string $code
	 * @return int
	 */
	public static function tifcode(string $code): int
	{
		return Arr::get(self::$tif_codes,$code);
	}

	public static function validateSignature(string $orig,string $pk,string $sig): bool
	{
		return (sodium_crypto_sign_open($sig.$orig, $pk) !== FALSE);
	}

	public static function validateSignatures(array $data_received,array $decode_request): bool
	{
		$data_to_validate = Arr::get($data_received,'client').Arr::get($data_received,'server');

		if (! self::validateSignature($data_to_validate,Arr::get($decode_request,'client.idk'),Arr::get($decode_request,'ids'))) {
			return FALSE;
		}

		if (Arr::get($decode_request,'urs') && Arr::get($decode_request,'client.vuk')) {
			if (Arr::get($decode_request,'client.pidk')) {
				$sqrl_pubkey = Pubkey::check(Arr::get($decode_request,'client.pidk'));

				if ($sqrl_pubkey && ! self::validateSignature($data_to_validate,$sqrl_pubkey->vuk,Arr::get($decode_request,'urs'))) {
					return FALSE;
				}

			} else if (! Arr::get($decode_request,'client.pidk')
				&& ! self::validateSignature($data_to_validate,Arr::get($decode_request,'client.vuk'),Arr::get($decode_request,'urs')))
			{
				return FALSE;
			}
		}

		if (Arr::get($decode_request,'pids')
			&& Arr::get($decode_request,'client.pidk')
			&& ! self::validateSignature($data_to_validate,Arr::get($decode_request,'client.pidk'),Arr::get($decode_request,'pids')))
		{
			return FALSE;
		}

		return TRUE;
	}
}
